/* Copyright 2008 Daniel Vokó <vokoda@gmail.com>
 * 
 * This file is part of Sagittarius.
 *
 * Sagittarius is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * Sagittarius is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Sagittarius.  If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using Gtk;
using Gdk;
using Sagittarius.Metadata;
using System.Collections.Generic;
using System.Threading;
using Sagittarius.Exiftool;

namespace Sagittarius
{
	public class MainClass
	{
		private PhotoInfo activePhoto;
		private List<PhotoInfo> selectedPhotos;
		private ListStore photoStore;
		private string selectedPath;
		private XmpMetadata metadata;
		public bool photosPresent;
		private int numOfPhotosInDirectory, numOfPhotosLoaded;
		private KeywordManager keywordManager;
		private CacheManager cacheManager;
		public Config config;
		
		public MainClass( )
		{
			config = new Config();
			config.ReadConfig();
		}

		public static void Main (string[] args)
		{
			Application.Init ();
			MainWindow win = new MainWindow ();
			win.Show ();
			Application.Run ();
		}

		/// <summary>
		/// Creates a ListStore of the image files in the given directory
		/// </summary>
		/// <param name="path">
		/// A <see cref="System.String"/>
		/// </param>
		/// <returns>
		/// A <see cref="Gtk.ListStore"/>
		/// </returns>
		public Gtk.ListStore ReadDirectoryContents(string path)
		{
			//Console.WriteLine("Entering MainClass.ReadDirectoryContents");
			if(cacheManager!=null)
			{
				cacheManager.WriteCacheForDirectory();
			}
			selectedPath=path;
			//cacheManager = new CacheManager(path,config);
			//Console.WriteLine(path);
			PopulateImageStore();
			//Console.WriteLine("Leaving MainClass.ReadDirectoryContents");
			return photoStore;
		}
		
		private void PopulateImageStore()
		{
			//Console.WriteLine("Entering Sagittarius.Main.PopulateImageStore");
			photoStore = new ListStore(typeof(PhotoInfo));
			photosPresent=false;
			string[] files = System.IO.Directory.GetFiles(selectedPath); // all the files in current folder
			numOfPhotosInDirectory = 0;
			this.numOfPhotosLoaded=0;
			cacheManager = new CacheManager(selectedPath,config);
			PhotoInfoFactory.KillThreads();
			PhotoInfoFactory factory = new PhotoInfoFactory(config,cacheManager);
			factory.OnPhotoInfoReady+=new PhotoInfoFactory.OnPhotoInfoReadyHandler (PhotoInfoIsAvailable);
			for(int k=0;k<files.Length;k++)
			{
				if(Misc.IsFileSupported(files[k])) // if current file is a supported image
				{
					numOfPhotosInDirectory++;
					photosPresent=true;
					factory.BeginProcessPhoto(files[k]);
					
				}
			}
			//Console.WriteLine("Leaving Sagittarius.Main.PopulateImageStore");
		}

		private void PhotoInfoIsAvailable(PhotoInfoFactory factory, PhotoInfoReadyEventArgs args)
		{
			photoStore.AppendValues(args.PInfo);
			this.numOfPhotosLoaded++;
			RaiseLoadSaveStatusUpdateEvent(this.numOfPhotosLoaded,numOfPhotosInDirectory,"Loading photos...");
			
		}
		
		public void DeselectImage()
		{
			metadata = null;
		}

		public void ActivatePhoto(TreePath path)
		{
			//Console.WriteLine("Entering Main.ActivateImage");
			TreeIter iter;
			photoStore.GetIter (out iter, path);
			activePhoto = (PhotoInfo) photoStore.GetValue(iter,0);
			//Console.WriteLine("Exiting Main.ActivateImage");
		}
		
		public void SelectImages(TreePath[] paths)
		{
			//Console.WriteLine("Entering Main.SelectImages");
			selectedPhotos=new List<PhotoInfo>();
			TreeIter[] iters = new TreeIter[paths.Length];
			for (int i=0; i<iters.Length;i++)
			{
				photoStore.GetIter (out iters[i],paths[i]);
				selectedPhotos.Add(((PhotoInfo) photoStore.GetValue(iters[i],0)));
			}
			GatherMetadata();
			//Console.WriteLine("Leaving Main.SelectImages");
		}
		
		/// <summary>
		/// This method gathers the metadata of:
		/// either the currently active image, if there is only one image selected;
		/// or all of the currently selected images
		/// </summary>
		private void GatherMetadata()
		{
			//Console.WriteLine("Entering Main.GatherMetadata");
			if (selectedPhotos.Count==1)
			{
				metadata = ((PhotoInfo)selectedPhotos[0]).Metadata;
			}
			else
			{
				metadata = new XmpMetadata();
				/*foreach (MetadataElement element in metadata.XmpExifSection.Elements)
				{
					element.Value = activeImage.Metadata.XmpExifSection.GetElement(element.Key).Value;
					for(int i=0; i<selectedImages.Count;i++)
					{
						if(element.Value!=((Image)selectedImages[i]).Metadata.XmpExifSection.GetElement(element.Key).Value)
						{
							element.Value=Strings.MULTIPLE_VALUES;
							break;
						}
					}
				}*/
				//Console.WriteLine("Gathering metadata...");
				foreach (MetadataElement element in metadata.XmpIptcSection.Elements)
				{
					element.Value = activePhoto.Metadata.XmpIptcSection.GetValue(element.Key);
					for(int i=0; i<selectedPhotos.Count;i++)
					{
						if(element.Value!=((PhotoInfo)selectedPhotos[i]).Metadata.XmpIptcSection.GetValue(element.Key))
						{
							element.Value=Misc.MULTIPLE_VALUES;
							break;
						}
					}
				}
				//keyword management
				Keywords[] keywordsArray= new Keywords[selectedPhotos.Count];
				for(int i=0; i<selectedPhotos.Count;i++)
				{
					keywordsArray[i]=((PhotoInfo)selectedPhotos[i]).Metadata.XmpIptcSection.Keywords;
				}
				keywordManager = new KeywordManager(keywordsArray);
				metadata.XmpIptcSection.SetElement(ExiftoolTags.IPTC_KEYWORDS,keywordManager.GenerateUnion());
			}
			//Console.WriteLine("Leaving Main.GatherMetadata");
		}
		
		public void UpdateIptc(MetadataSection newIptcSection)
		{
			bool keywordsChanged=false;
			foreach(MetadataElement newElement in newIptcSection.Elements)
			{
				if (newElement.Key==Exiftool.ExiftoolTags.IPTC_KEYWORDS)
				{
					keywordsChanged=true;
				}
				for(int i = 0; i<selectedPhotos.Count;i++)
				{
					foreach(MetadataElement oldElement in ((PhotoInfo)selectedPhotos[i]).Metadata.XmpIptcSection.Elements)
					{
						if(oldElement.Key==newElement.Key)
						{
							oldElement.Value=newElement.Value;
							metadata.XmpIptcSection.SetElement(newElement.Key,newElement.Value);
							break;
						}
					}
					((PhotoInfo)selectedPhotos[i]).MetadataChanged=true;
				}
			}
			if((selectedPhotos.Count>=2)&&keywordsChanged)
			{
				keywordManager.SetKeywords(newIptcSection.GetValue(ExiftoolTags.IPTC_KEYWORDS));
				foreach(PhotoInfo image in selectedPhotos)
				{
					image.Metadata.XmpIptcSection.UpdateKeywords();
				}
			}
		}
		
		public void SaveMetadata()
		{
			int i=0;
			foreach(object[] image in photoStore)
			{
				if(((PhotoInfo)image[0]).MetadataChanged)
				{
					//((Image)image[0]).WriteMetadata(config.CreateBackup);
					((PhotoInfo)image[0]).WriteMetadata();
					((PhotoInfo)image[0]).MetadataChanged=false;
				}
				i++;
				while(Gtk.Application.EventsPending())
				{
					Gtk.Application.RunIteration();
				}
				RaiseLoadSaveStatusUpdateEvent(i,numOfPhotosInDirectory,"Saving metadata to files...");
			}
		}

		public void DiscardChanges()
		{
			//Console.WriteLine("Entering Main.DiscardChanges");
			foreach(object[] image in photoStore)
			{
				if(((PhotoInfo)image[0]).MetadataChanged)
				{
					((PhotoInfo)image[0]).DiscardMetadataChanges();
				}
			}
			GatherMetadata();
			//Console.WriteLine("Leaving Main.DiscardChanges");
		}

		/// <summary>
		/// Extracts the preview of the selected image, on the specified size
		/// </summary>
		/// <param name="width">
		/// A <see cref="System.Int32"/>
		/// </param>
		/// <param name="height">
		/// A <see cref="System.Int32"/>
		/// </param>
		/// <returns>
		/// A <see cref="Gdk.Pixbuf"/>
		/// </returns>
		public Gdk.Pixbuf GetActiveImagePreview(int width, int height)
		{
			//Console.WriteLine("Entering MainClass.GetPreview");
			//Console.WriteLine("Leaving MainClass.GetPreview");
			activePhoto.OnPreviewLoaded+=new PhotoInfo.OnPreviewLoadedHandler(ActivePhotoPreviewLoaded);
			return activePhoto.GetPreview(width,height);
		}
		
		private void ActivePhotoPreviewLoaded(PhotoInfo o, Sagittarius.PreviewLoadedEventArgs args)
		{
			//Console.WriteLine("Entering Sagittarius.MainClass.ActivePhotoPreviewLoaded");
			if(args.Filename==this.ActivePhoto.Filename)
			{
				this.RaisePreviewChangedEvent(args.Preview);
			}
			//Console.WriteLine("Leaving Sagittarius.MainClass.ActivePhotoPreviewLoaded");
		}
		
		public bool CheckUnsavedChanges()
		{
			//Console.WriteLine("Entering Main.CheckUnsavedChanges");
			bool result=false;
			if(photoStore!=null)
			{
				foreach(object[] image in photoStore)
				{
					if(((PhotoInfo)image[0]).MetadataChanged)
					{
						result = true;
					}
				}
			}
			//Console.WriteLine("Leaving Main.CheckUnsavedChanges");
			return result;
		}
		
		public void Quit()
		{
		}
		
		
#region events

		public event LoadSaveStatusUpdateHandler OnLoadSaveStatusUpdate;
		
		public delegate void LoadSaveStatusUpdateHandler (object sender, LoadSaveStatusUpdateEventArgs args);
		
		private void RaiseLoadSaveStatusUpdateEvent (int act, int max, string message)
		{
			//Console.WriteLine("act: "+act+ "max: "+max);
			if (OnLoadSaveStatusUpdate!=null)
			{
				while(Gtk.Application.EventsPending())
				{
					Gtk.Application.RunIteration();
				}
				OnLoadSaveStatusUpdate(this,new LoadSaveStatusUpdateEventArgs(act, max, message));
			}
		}
		
		public event OnPreviewChangedHandler OnPreviewChanged;

		public delegate void OnPreviewChangedHandler (Sagittarius.MainClass o, PreviewLoadedEventArgs args);
		
		private void RaisePreviewChangedEvent (Pixbuf preview)
		{
			if (OnPreviewChanged!=null)
			{
				OnPreviewChanged(this,new Sagittarius.PreviewLoadedEventArgs(preview,this.ActivePhoto.Filename));
			}
		}

#endregion

#region property

		public ListStore ImageStore
		{
			get { return photoStore; }
		}
		public PhotoInfo ActivePhoto
		{
			get { return activePhoto; }
			set { activePhoto = value; }
		}
		public XmpMetadata Metadata
		{
			get { return metadata; }
		}
		public Config Config
		{
			get { return config;}
		}
		
#endregion
		
		
		
	}
}