/* Copyright 2008 Daniel Vokó <vokoda@gmail.com>
 * 
 * This file is part of Sagittarius.
 *
 * Sagittarius is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * Sagittarius is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Sagittarius.  If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using Gtk;
using Sagittarius.Metadata;
using System.Collections.Generic;
using Gdk;

namespace Sagittarius
{
	public partial class MainWindow: Gtk.Window
	{	
		private FileSystemReader fsReader; // an object responsible for the data of the DirectoryTree;
		private BookmarkManager bmManager;
		private FileSystemReader bookmarkReader;
		private int oldPrevWidth, oldPrevHeight; // size of the imageWidget (before resizing)
		private MainClass main;
		
		public MainWindow (): base (Gtk.WindowType.Toplevel)
		{
			Build ();
			
			SetEditBtnsStatus(false);
			SetSaveDiscardBtnsStatus(false);
			
			main = new MainClass(/*this*/);
			
			// theese variables are used during preview resizing
			oldPrevWidth= previewWidget.Allocation.Width;
			oldPrevHeight = previewWidget.Allocation.Height;

			bmManager = new BookmarkManager();

			SetupFileSystemTreeView(); // initialization of the Directory Tree
			SetupBookmarkTreeView();
			SetupThumbView(); // initialization of the Thumbnail-view
			SetupMetadataNotebook(); // initialization of the metadata notebook widget
			main.OnLoadSaveStatusUpdate+=new MainClass.LoadSaveStatusUpdateHandler(LoadSaveStatusUpdated);
			main.OnPreviewChanged+= new MainClass.OnPreviewChangedHandler(UpdatePreview);
		}
		
#region FileSystemTreeView methods
		
		/// <summary>
		/// Initialization of the FilesystemReader 
		/// and the treeview widget displaying the filesystem
		/// </summary>
		private void SetupFileSystemTreeView()
		{
			fsReader = new FileSystemReader(false);
			
			TreeViewColumn iconCol = new TreeViewColumn();
			iconCol.Sizing=TreeViewColumnSizing.Autosize;
			CellRendererPixbuf iconCell = new CellRendererPixbuf();
			iconCol.PackStart(iconCell,false);
			
			CellRendererText nameCell = new CellRendererText();
			iconCol.PackStart(nameCell,true);
			
			fsTreeView.AppendColumn(iconCol);
			
			iconCol.SetCellDataFunc(iconCell,new Gtk.TreeCellDataFunc(RenderFsTreeViewDirIcon));
			iconCol.SetCellDataFunc(nameCell,new Gtk.TreeCellDataFunc(RenderFsTreeViewDirName));
			
			fsTreeView.Model=fsReader.DirectoryTree;
			fsTreeView.HeadersVisible=false;
		}
		
		private void RenderFsTreeViewDirName(Gtk.TreeViewColumn column, CellRenderer cell, TreeModel model, TreeIter iter)
		{
			DirInfo dirInfo = (DirInfo) model.GetValue(iter,0);
			(cell as Gtk.CellRendererText).Text = dirInfo.Name;
		}
		
		private void RenderFsTreeViewDirIcon(Gtk.TreeViewColumn column, CellRenderer cell, TreeModel model, TreeIter iter)
		{
			DirInfo dirInfo = (DirInfo) model.GetValue(iter,0);
			(cell as Gtk.CellRendererPixbuf).Pixbuf = dirInfo.Icon;
		}
		
		protected virtual void OnFsTreeViewRowExpanded (object o, Gtk.RowExpandedArgs args)
		{
			//Console.WriteLine("Entering OnFsTreeViewRowExpanded");
			fsReader.ExploreDir(args.Path);
			//Console.WriteLine("Exiting OnFsTreeViewRowExpanded");
		}

		protected virtual void OnFsTreeViewRowCollapsed (object o, Gtk.RowCollapsedArgs args)
		{
			fsReader.UnexploreDir(args.Path);
		}

		protected virtual void OnFsTreeViewRowActivated (object o, Gtk.RowActivatedArgs args)
		{
			//Console.WriteLine("Entering OnFsTreeViewRowActivated");
			// when a row is activated, the contents of the directory it represents
			// is to be read
			// but first, all the models of the treeviews are to be recreated, 
			// to discard the contents of previously selected directory/images
			main.DeselectImage();
			CheckUnsavedChanges();
			DiscardPreview();
			DiscardThumbView();
			DiscardMetadataView();
			SetEditBtnsStatus(false);
			PopulateThumbView(main.ReadDirectoryContents(fsReader.GetDirName(args.Path)));
			//Console.WriteLine("Exiting OnFsTreeViewRowActivated");
		}

#endregion
		
#region Bookmark-treeview methods
		
		private void SetupBookmarkTreeView()
		{
			//Console.WriteLine("Entering MainWindow.SetupBookmarkTreeView");
			TreeViewColumn iconCol = new TreeViewColumn();
			iconCol.Sizing=TreeViewColumnSizing.Autosize;
			CellRendererPixbuf iconCell = new CellRendererPixbuf();
			iconCol.PackStart(iconCell,false);
			
			CellRendererText nameCell = new CellRendererText();
			iconCol.PackStart(nameCell,true);
			
			bookmarkTreeView.AppendColumn(iconCol);
			
			iconCol.SetCellDataFunc(iconCell,new Gtk.TreeCellDataFunc(RenderBookmarkTreeViewDirIcon));
			iconCol.SetCellDataFunc(nameCell,new Gtk.TreeCellDataFunc(RenderBookmarkTreeViewDirName));
			
			//bookmarkTreeView.HeadersVisible=false;
			
			bookmarkReader = new FileSystemReader(false,bmManager.BookmarkList);
			bookmarkTreeView.Model=bookmarkReader.DirectoryTree;
			//Console.WriteLine("Leaving MainWindow.SetupBookmarkTreeView");
		}
		
		private void RenderBookmarkTreeViewDirName(Gtk.TreeViewColumn column, CellRenderer cell, TreeModel model, TreeIter iter)
		{
			DirInfo dirInfo = (DirInfo) model.GetValue(iter,0);
			(cell as Gtk.CellRendererText).Text = dirInfo.Name;
		}
		
		private void RenderBookmarkTreeViewDirIcon(Gtk.TreeViewColumn column, CellRenderer cell, TreeModel model, TreeIter iter)
		{
			DirInfo dirInfo = (DirInfo) model.GetValue(iter,0);
			(cell as Gtk.CellRendererPixbuf).Pixbuf = dirInfo.Icon;
		}

		protected virtual void OnAddBookmarkBtnClicked (object sender, System.EventArgs e)
		{
			//Console.WriteLine("Entering MainWindow.OnAddBookmarkBtnClicked");
			TreePath[] path =fsTreeView.Selection.GetSelectedRows();
			if(path.Length>0)
			{
				bmManager.AddBookmark(fsReader.GetDirName(path[0]));
				bmManager.SaveBookmarks(false);
				bookmarkReader.AddPath(fsReader.GetDirName(path[0]));
			}
			//Console.WriteLine("Leaving MainWindow.OnAddBookmarkBtnClicked");
		}

		protected virtual void OnRemoveBookmarkBtnClicked (object sender, System.EventArgs e)
		{
			TreePath[] path =bookmarkTreeView.Selection.GetSelectedRows();
			if (path.Length>0)
			{
				for (int i= path[0].Depth;i>1;i--)
				{
					path[0].Up();
				}
				bmManager.DeleteBookmark(bookmarkReader.GetDirName(path[0]));
				bmManager.SaveBookmarks(false);
				bookmarkReader.DeletePath(path[0]);
			}
		}

		protected virtual void OnBookmarkTreeViewRowActivated (object o, Gtk.RowActivatedArgs args)
		{
			//Console.WriteLine("Entering MainWindow.OnBookmarkTreeViewRowActivated");
			// when a row is activated, the contents of the directory it represents
			// is to be read
			// but first, all the models of the treeviews are to be recreated, 
			// to discard the contents of previously selected directory/images
			main.DeselectImage();
			CheckUnsavedChanges();
			DiscardPreview();
			DiscardThumbView();
			DiscardMetadataView();
			SetEditBtnsStatus(false);
			PopulateThumbView(main.ReadDirectoryContents(bookmarkReader.GetDirName(args.Path)));
			//Console.WriteLine("Leaving MainWindow.OnBookmarkTreeViewRowActivated");
		}
		
		protected virtual void OnBookmarkTreeViewRowExpanded (object o, Gtk.RowExpandedArgs args)
		{
			bookmarkReader.ExploreDir(args.Path);
		}

		protected virtual void OnBookmarkTreeViewRowCollapsed (object o, Gtk.RowCollapsedArgs args)
		{
			bookmarkReader.UnexploreDir(args.Path);
		}

#endregion

#region ThumbView methods

		private void SetupThumbView()
		{
			//Console.WriteLine("Entering MainWindow.SetupThumbView");
			TreeViewColumn thumbCol = new TreeViewColumn();
			CellRendererPixbuf thumbCell = new CellRendererPixbuf();
			thumbCol.AddAttribute(thumbCell, "pixbuf", 0);  // it will display an image on the left (the image thumb) ....
			thumbCol.PackStart(thumbCell, false);
			thumbCol.SetCellDataFunc(thumbCell, new TreeCellDataFunc(RenderThumbnail));
			thumbView.AppendColumn (thumbCol);
		
			TreeViewColumn markupCol = new TreeViewColumn ();
			CellRenderer markupCell = new CellRendererText();
			markupCol.AddAttribute (markupCell, "markup", 1);  // ...and a lot of formatted information
			markupCol.PackStart(markupCell, true); 
			markupCol.SetCellDataFunc(markupCell, new TreeCellDataFunc(RenderMarkup));
			thumbView.AppendColumn (markupCol);
			
			thumbView.HeadersVisible=false;
			thumbView.Selection.Mode=SelectionMode.Multiple;
			thumbView.Selection.Changed +=  new EventHandler(OnThumbViewSelectionChanged);
			//Console.WriteLine("Leaving MainWindow.SetupThumbView");
		}

		private void RenderThumbnail (Gtk.TreeViewColumn columnt,Gtk.CellRenderer cell, Gtk.TreeModel model,Gtk.TreeIter iter)
		{
			PhotoInfo currentPInfo = (PhotoInfo) model.GetValue(iter,0);
			((CellRendererPixbuf) cell).Pixbuf = currentPInfo.Thumbnail;
		}

		private void RenderMarkup (Gtk.TreeViewColumn columnt,Gtk.CellRenderer cell, Gtk.TreeModel model,Gtk.TreeIter iter)
		{
			PhotoInfo currentPInfo = (PhotoInfo) model.GetValue(iter,0);
			((CellRendererText) cell).Markup = currentPInfo.Markup;
		}

		private void PopulateThumbView(ListStore pInfoStore)
		{
			//Console.WriteLine("Entering MainWindow.PopulateThumbView");
			thumbView.Model=pInfoStore;
			thumbView.ColumnsAutosize();
			//Console.WriteLine("Exiting MainWindow.PopulateThumbView");
		}
		private void DiscardThumbView()
		{
			thumbView.Model=null;
		}
		
		
		private void OnThumbViewSelectionChanged (object o, EventArgs args)
		{
			//Console.WriteLine("Entering OnThumbViewSelectionChanged");
			// This is to be called, because SelectionChanged events
			// occur more frequently and what's more important, before the Cursorchanged event
			// and as this event triggers the Main.SelectImages method,
			// at multiple selection scenarios it can lead to bad behaviour
			ActivateImageAtCursor();
			
			TreeModel model;
 			TreePath[] selectedPaths = ((TreeSelection)o).GetSelectedRows(out model);
 			if (selectedPaths.Length>0)
 			{
 				main.SelectImages(selectedPaths);
 			}
 			ShowMetadata();
			//Console.WriteLine("Leaving OnThumbViewSelectionChanged");
		}
		
		protected virtual void OnThumbViewCursorChanged (object sender, System.EventArgs e)
		{
			//Console.WriteLine("Entering OnThumbViewCursorChanged");
			ActivateImageAtCursor();
			// displaying metadata on notebook
			ShowMetadata();
			ShowPreview();
			//Console.WriteLine("Leaving OnThumbViewCursorChanged");
		}
		
		private void ActivateImageAtCursor()
		{
			TreePath path;
			TreeViewColumn col;
			thumbView.GetCursor(out path, out col);
			if(path!=null)
			{
				main.ActivatePhoto(path);
				SetEditBtnsStatus(true);
			}
			else
			{
				SetEditBtnsStatus(false);
			}
		}

#endregion

#region MetadataNotebook methods

		private void SetupMetadataNotebook()
		{
			//Console.WriteLine("Entering MainWindow.SetupMetadataNotebook");
			TreeViewColumn keyCol = new TreeViewColumn();
			keyCol.Title="Key";
			CellRendererText keyCell = new CellRendererText();
			keyCol.PackStart(keyCell,true);
			
			TreeViewColumn valueCol = new TreeViewColumn();
			valueCol.Title = "Value";
			CellRendererText valueCell = new CellRendererText();
			valueCol.PackStart(valueCell,true);

			exifTreeView.AppendColumn(keyCol);
			exifTreeView.AppendColumn(valueCol);

			keyCol.AddAttribute(keyCell,"text",0);
			valueCol.AddAttribute(valueCell,"text",1);

			keyCol = new TreeViewColumn();
			keyCol.Title="Key";
			keyCell = new CellRendererText();
			keyCol.PackStart(keyCell,true);
			
			valueCol = new TreeViewColumn();
			valueCol.Title = "Value";
			valueCell = new CellRendererText();
			valueCol.PackStart(valueCell,true);

			iptcTreeView.AppendColumn(keyCol);
			iptcTreeView.AppendColumn(valueCol);

			keyCol.AddAttribute(keyCell,"text",0);
			valueCol.AddAttribute(valueCell,"text",1);
			//Console.WriteLine("Leaving MainWindow.SetupMetadataNotebook");
		}
		
		public void ShowMetadata()
		{
			//Console.WriteLine("Entering Sagittarius.ShowMetadata");
			if (main.Metadata!=null)
			{
				//Console.WriteLine("metadata is not null");
				MetadataSection exifSect = main.Metadata.XmpExifSection;
				MetadataSection iptcSect = main.Metadata.XmpIptcSection; 
				
				ListStore exifLS = new ListStore(typeof (string), typeof (string));
				for(int j = 0; j< exifSect.Length; j++)
				{
					exifLS.AppendValues(exifSect.Elements[j].Key,exifSect.Elements[j].Value);
				}
				exifTreeView.Model=exifLS;
				
				ListStore iptcLS = new ListStore(typeof (string), typeof (string));
				for(int j = 0; j< iptcSect.Length; j++)
				{
					iptcLS.AppendValues(iptcSect.Elements[j].Key,iptcSect.Elements[j].Value);
				}
				iptcTreeView.Model=iptcLS;
			}
			//Console.WriteLine("Leaving Sagittarius.ShowMetadata");
		}
		private void DiscardMetadataView()
		{
			//exifTreeView.Model=null;
			iptcTreeView.Model=null;
		}

#endregion

#region PreviewWidget methods

		private void ShowPreview()
		{
			if(oldPrevWidth>main.Config.PrevMaxSize)
			{
				oldPrevWidth= main.config.PrevMaxSize;
			}
			if(oldPrevHeight>main.Config.PrevMaxSize)
			{
				oldPrevHeight= main.config.PrevMaxSize;
			}
			previewWidget.Pixbuf = main.GetActiveImagePreview(oldPrevWidth, oldPrevHeight);
		}
		
		private void UpdatePreview(Sagittarius.MainClass o, Sagittarius.PreviewLoadedEventArgs args)
		{
			//Console.WriteLine("Entering Sagittarius.MainWindow.UpdatePreview");
			previewWidget.Pixbuf=args.Preview;
			//Console.WriteLine("Leaving Sagittarius.MainWindow.UpdatePreview");
		}
		
		private void DiscardPreview()
		{
			previewWidget.Pixbuf=new Pixbuf(null,"null.png");
		}

		protected virtual void OnPreviewWidgetExposeEvent (object o, Gtk.ExposeEventArgs args)
		{
			//Console.WriteLine("Entering MainWindow.OnImageWidgetExpose");
			if (previewWidget.Pixbuf!=null)
			{
				if ((oldPrevWidth != previewWidget.Allocation.Width) || (oldPrevHeight != previewWidget.Allocation.Height))
				{
					oldPrevWidth = previewWidget.Allocation.Width;
					oldPrevHeight = previewWidget.Allocation.Height;
					ShowPreview();
				}
			}
			//Console.WriteLine("Leaving MainWindow.OnImageWidgetExpose");
		}

#endregion

#region toolbar methods

		protected virtual void OnToolbEditIptcBtnActivated (object sender, System.EventArgs e)
		{
			Sagittarius.EditIptcDialog editIptcDialog = new EditIptcDialog(main.Metadata.XmpIptcSection);
			editIptcDialog.ShowAll();
			if (editIptcDialog.Run()==(int)ResponseType.Accept)
			{
				main.UpdateIptc(editIptcDialog.NewIptcSection);
				ShowMetadata();
				if (main.CheckUnsavedChanges())
				{
					SetSaveDiscardBtnsStatus(true);
				}
			}
			editIptcDialog.Hide();
			editIptcDialog.Destroy();
		}

		protected virtual void OnToolbarSaveBtnActivated (object sender, System.EventArgs e)
		{
			main.SaveMetadata();
			SetSaveDiscardBtnsStatus(false);
		}

		protected virtual void OnToolbarDiscardBtnActivated (object sender, System.EventArgs e)
		{
			main.DiscardChanges();
			SetSaveDiscardBtnsStatus(false);
			ShowMetadata();
		}
		
		private void SetEditBtnsStatus(bool sensitive)
		{
			toolbarEditExifBtn.Sensitive=sensitive;
			toolbarEditIptcBtn.Sensitive=sensitive;
		}
		
		private void SetSaveDiscardBtnsStatus(bool sensitive)
		{
			toolbarSaveBtn.Sensitive=sensitive;
			toolbarDiscardBtn.Sensitive=sensitive;
		}
		
		private void Quit()
		{
			CheckUnsavedChanges();
			main.Quit();
			PhotoInfoFactory.KillThreads();
			Application.Quit ();
		}

#endregion
		
#region eventhandlers

		protected void OnDeleteEvent (object sender, DeleteEventArgs a)
		{
			Quit();
			a.RetVal = true;
		}

		protected virtual void OnMenuFileQuitActivated (object sender, System.EventArgs e)
		{
			Quit();
		}

		protected virtual void OnMenuEditPreferencesActivated (object sender, System.EventArgs e)
		{
			PreferencesDialog preferencesDialog = new PreferencesDialog(main.config);
			preferencesDialog.ShowAll();
			preferencesDialog.Run();
			preferencesDialog.Hide();
			preferencesDialog.Destroy();
		}
		
		protected virtual void OnAboutActivated (object sender, System.EventArgs e)
		{
			AboutBox aboutBox = new AboutBox();
			aboutBox.Show();
		}

#endregion

		private void LoadSaveStatusUpdated (object sender, LoadSaveStatusUpdateEventArgs e)
		{
			//Console.WriteLine("Entering eventhandler");
			thumbStatusBox.Visible=true;
			progressbar.Adjustment.Lower=0;
			progressbar.Adjustment.Upper=e.max;
			progressbar.Adjustment.Value=e.act;
			progressbar.Text=Math.Round(((double)e.act/(double)e.max)*100).ToString()+"%";
			statusLabel.Text=e.message;
			if(e.max==e.act)
			{
				statusLabel.Text="";
				thumbStatusBox.Visible=false;
			}
		}
		
		private void CheckUnsavedChanges()
		{
			if (main.CheckUnsavedChanges())
			{
				MessageDialog md = new MessageDialog(
					this,DialogFlags.Modal,
					MessageType.Warning,
					ButtonsType.YesNo,
					"There are changes not saved yet! Do you want to save them?");
				ResponseType result = (ResponseType)md.Run();
				if(result == ResponseType.Yes)
				{
					main.SaveMetadata();
					SetSaveDiscardBtnsStatus(false);
				}
				else
				{
					main.DiscardChanges();
					SetSaveDiscardBtnsStatus(false);
				}
				md.Hide();
				md.Destroy();
			}
		}

		protected virtual void OnRenameActionActivated (object sender, System.EventArgs e)
		{
			RenameDialog renDi = new RenameDialog();
			renDi.Show();
		}


	}
}